from pathlib import Path

import toml
from pydantic import ValidationError

from delfino.constants import PYPROJECT_TOML_FILENAME
from delfino.models import PyprojectToml

_RC_FILE_NAME = ".delfinorc"


class ConfigValidationError(ValueError):
    pass


def _rc_locations(project_root: Path) -> list[Path]:  # pragma: no cover
    return [
        Path.home() / _RC_FILE_NAME,
        project_root / PYPROJECT_TOML_FILENAME,
        project_root / _RC_FILE_NAME,
    ]


def load_config(project_root: Path) -> PyprojectToml:
    """Loads a config with defined precedence.

    Looks up the following locations in this order (later overwrite earlier):
        - $USER/.delfinorc
        - <project_root>/pyproject.toml
        - <project_root>/.delfinorc

    The ``.delfinorc`` files have the same TOML format and structure as the
    ``pyproject.toml`` file.

    Args:
        project_root: Root of the project.
    """
    pyproject_toml = PyprojectToml()

    for rc_file in _rc_locations(project_root):
        if not rc_file.is_file():
            continue
        try:
            pyproject_toml = PyprojectToml(**toml.load(rc_file))
        except ValidationError as exc:
            raise ConfigValidationError(f"Delfino appears to be misconfigured in '{rc_file}': {exc}") from exc

    return pyproject_toml



import logging
import os
import sys
from pathlib import Path
from typing import Any, Optional

import click

from delfino.click_utils.command import CommandRegistry
from delfino.config import ConfigValidationError, load_config
from delfino.constants import ENTRY_POINT, PYPROJECT_TOML_FILENAME
from delfino.internal_parameters.completion import (
    install_completion_option,
    show_completion_option,
)
from delfino.internal_parameters.help import extended_help_option
from delfino.internal_parameters.verbosity import log_level_option
from delfino.models.app_context import AppContext
from delfino.models.pyproject_toml import PyprojectToml
from delfino.utils import get_package_manager


class Commands(click.Group):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        sys.path.append(os.getcwd())

        self._project_root = Path(os.getcwd())

        # When evaluating available commands, the program should not fail. We save the exception
        # to show it only when a command is executed.
        self._pyproject_toml_validation_error: Optional[ConfigValidationError] = None

        try:
            self._pyproject_toml = load_config(self._project_root)
        except ConfigValidationError as exc:
            self._pyproject_toml = PyprojectToml()
            self._pyproject_toml_validation_error = exc

        self._command_registry = CommandRegistry(
            plugins_configs=self._pyproject_toml.tool.delfino.plugins,
            local_command_folders=self._pyproject_toml.tool.delfino.local_command_folders,
        )

    def list_commands(self, ctx: click.Context) -> list[str]:
        """Override as MultiCommand always returns []."""
        del ctx
        return sorted(command.name for command in self._command_registry.visible_commands)

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        """Override to give all commands a common ``AppContext`` or fail if ``pyproject.toml`` is broken/missing."""
        if (cmd := self._command_registry.get(cmd_name, None)) is None:
            return None  # command doesn't exist

        if ctx.resilient_parsing:  # do not fail on auto-completion
            return cmd.command

        if self._pyproject_toml_validation_error:
            click.secho(str(self._pyproject_toml_validation_error), fg="red", err=True)
            raise click.Abort() from self._pyproject_toml_validation_error

        ctx.obj = AppContext(
            project_root=self._project_root,
            pyproject_toml=self._pyproject_toml,
            package_manager=get_package_manager(self._project_root, self._pyproject_toml),
            plugin_config=cmd.package.plugin_config,
        )

        return cmd.command

    def invoke(self, ctx: click.Context) -> Any:
        """Override to turn ``AssertionError`` exception into ``click.exceptions.Exit``."""
        try:
            return super().invoke(ctx)
        except AssertionError as exc:
            click.secho(
                f"Command '{ctx.invoked_subcommand}' is misconfigured. {exc}",
                fg="red",
                err=True,
            )
            raise click.exceptions.Exit(1)

    def get_help(self, *args, **kwargs) -> str:
        help_str = super().get_help(*args, **kwargs)

        if self._command_registry.hidden_commands:
            if logging.root.level == logging.DEBUG:
                help_str += click.style(
                    f"\n\nDisabled command{'s' if len(self._command_registry.hidden_commands) > 1 else ''}: "
                    + ", ".join(
                        sorted(
                            [
                                f"{command.package.plugin_name}/{command.command.name}"
                                for command in self._command_registry.hidden_commands
                            ]
                        )
                    )
                    + f" (see 'tool.{ENTRY_POINT}.<PLUGIN NAME>.disable_commands' in '{PYPROJECT_TOML_FILENAME}')",
                    fg="white",
                )

        return help_str


@click.group(cls=Commands)
@extended_help_option
@click.version_option()
@log_level_option
@show_completion_option
@install_completion_option
def main(log_level=None):
    del log_level


if __name__ == "__main__":
    main()



from delfino.decorators.files_folders import files_folders_option
from delfino.decorators.pass_app_context import pass_app_context
from delfino.decorators.pass_args import pass_args

__all__ = ["files_folders_option", "pass_app_context", "pass_args"]



from enum import Enum
from pathlib import Path
from typing import Final

ENTRY_POINT: Final[str] = "delfino"
PYPROJECT_TOML_FILENAME: Final[str] = "pyproject.toml"
DEFAULT_LOCAL_COMMAND_FOLDERS: Final[tuple[Path, ...]] = (Path("commands"),)


class PackageManager(Enum):
    POETRY = "poetry"
    PIPENV = "pipenv"
    UV = "uv"
    UNKNOWN = "unknown"



from importlib import metadata
from subprocess import PIPE

from delfino.constants import PYPROJECT_TOML_FILENAME, PackageManager
from delfino.execution import OnError, run


def pyproject_toml_key_missing(key: str) -> str:
    return f"Key '{key}' is missing from the '{PYPROJECT_TOML_FILENAME}' file."


def pip_package_installed(name: str, sub_process: bool = False) -> bool:
    if sub_process:
        return (
            run(
                f"pip show -q {name}",
                stdout=PIPE,
                stderr=PIPE,
                on_error=OnError.PASS,
                shell=True,
            ).returncode
            == 0
        )

    try:
        metadata.Distribution.from_name(name)
    except metadata.PackageNotFoundError:
        return False

    return True


def assert_pip_package_installed(name: str, required_by: str = "this command") -> None:
    assert pip_package_installed(name), (
        f"Optional Python package '{name}' is required by {required_by} but not installed."
    )


def assert_package_manager_is_known(package_manager: PackageManager) -> None:
    assert package_manager != PackageManager.UNKNOWN, (
        "Cannot determine package manager used in this project. Only the following ones are supported: "
        + ", ".join(member.value for member in PackageManager.__members__.values() if member != PackageManager.UNKNOWN)
    )



from delfino.models.app_context import AppContext
from delfino.models.pyproject_toml import (
    Delfino,
    PluginConfig,
    Poetry,
    PyprojectToml,
    Tool,
)

__all__ = ["AppContext", "Delfino", "PluginConfig", "Poetry", "PyprojectToml", "Tool"]



from collections import ChainMap
from logging import getLogger
from typing import cast

import click

from delfino.click_utils.command import get_root_command
from delfino.decorators.files_folders import FILES_FOLDERS_OPTION_CALLBACK
from delfino.decorators.pass_args import PASS_ARGS_CALLBACK
from delfino.models.app_context import AppContext

_LOG = getLogger(__name__)


def get_command_groups(app_context: AppContext) -> dict[str, list[str]]:
    return {
        **app_context.plugin_config.command_groups,
        **app_context.pyproject_toml.tool.delfino.command_groups,
    }


def _get_target_command_names(group_name: str, app_context: AppContext) -> list[str]:
    if (target_command_names := get_command_groups(app_context).get(group_name, None)) is None:
        raise click.exceptions.Abort(f"Command group '{group_name}' does not exist.")

    if not target_command_names:
        _LOG.warning(f"Command group '{group_name}' contains no commands.")
        raise click.exceptions.Exit()

    return target_command_names


def execute_commands_group(group_name: str, click_context: click.Context, app_context: AppContext, **kwargs):
    target_command_names = _get_target_command_names(group_name, app_context)
    root_command = get_root_command(click_context)

    commands: dict[str, click.Command] = {
        command: cast(click.Command, root_command.get_command(click_context, command))
        for command in root_command.list_commands(click_context)
    }

    for target_name in target_command_names:
        if target_name not in commands:
            _LOG.warning(f"Command '{target_name}' from the '{group_name}' command group does not exist. Skipping.")
            continue

        if target_name in app_context.plugin_config.disable_commands:
            _LOG.debug(f"Skipping disabled command '{target_name}'.")
            continue

        command = commands[target_name]

        parameter_from_config = ChainMap(
            *(
                callback.parameter_from_config_in_group(click_context, command)
                for callback in [PASS_ARGS_CALLBACK, FILES_FOLDERS_OPTION_CALLBACK]
            )
        )

        click_context.forward(command, **kwargs, **parameter_from_config)



from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field

from delfino.constants import DEFAULT_LOCAL_COMMAND_FOLDERS


class PluginConfig(BaseModel):
    enable_commands: set[str] = Field(default_factory=set)
    disable_commands: set[str] = Field(default_factory=set)
    command_groups: dict[str, list[str]] = Field(default_factory=dict)

    @classmethod
    def empty(cls):
        return cls()

    model_config = ConfigDict(extra="allow")


class Delfino(BaseModel):
    local_command_folders: tuple[Path, ...] = DEFAULT_LOCAL_COMMAND_FOLDERS
    plugins: dict[str, PluginConfig] = Field(default_factory=dict)
    command_groups: dict[str, list[str]] = Field(default_factory=dict)
    model_config = ConfigDict(extra="allow")


class Poetry(BaseModel):
    name: str
    version: str
    scripts: dict[str, str] = Field(default_factory=dict)
    dependencies: dict[str, Any] = Field(default_factory=dict)
    model_config = ConfigDict(extra="allow")


class BuildSystem(BaseModel):
    requires: list[str] = Field(default_factory=list)
    build_backend: Optional[str] = None
    model_config = ConfigDict(extra="allow")


class Tool(BaseModel):
    poetry: Optional[Poetry] = None
    uv: Optional[dict[str, Any]] = None
    delfino: Delfino = Field(default_factory=Delfino)
    model_config = ConfigDict(populate_by_name=True)


class Project(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    dependencies: dict[str, Any] = Field(default_factory=dict)
    model_config = ConfigDict(extra="allow")


class PyprojectToml(BaseModel):
    tool: Tool = Field(default_factory=Tool)
    project: Optional[Project] = None
    build_system: Optional[BuildSystem] = None
    model_config = ConfigDict(extra="allow")



from pathlib import Path
from typing import Union

from delfino.constants import PackageManager
from delfino.models.pyproject_toml import PyprojectToml

ArgsList = list[Union[str, bytes, Path]]
ArgsType = Union[str, bytes, list]


def get_package_manager(project_root: Path, pyproject_toml: PyprojectToml) -> PackageManager:
    # Check build-system requires for poetry and uv
    if pyproject_toml.tool.poetry is not None or (project_root / "poetry.lock").exists():
        return PackageManager.POETRY

    if pyproject_toml.tool.uv is not None or (project_root / "uv.lock").exists():
        return PackageManager.UV

    if pyproject_toml.build_system and pyproject_toml.build_system.requires:
        requires = pyproject_toml.build_system.requires
        if any("poetry" in req for req in requires):
            return PackageManager.POETRY
        if any("hatchling" in req for req in requires):
            return PackageManager.UV

    if (project_root / "Pipfile").exists() or (project_root / "Pipfile.lock").exists():
        return PackageManager.PIPENV

    return PackageManager.UNKNOWN
